# yellowbrick.contrib.wrapper
# Wrapper for third-party estimators that implement the sklearn API
#
# Author:   Benjamin Bengfort
# Created:  Fri Oct 02 14:47:54 2020 -0400
#
# Copyright (C) 2020 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: wrapper.py [] benjamin@bengfort.com $

"""
Wrapper for third-party estimators that implement the sklearn API but do not directly
subclass the ``sklearn.base.BaseEstimator`` class. This method is a quick way to get
other estimators into Yellowbrick, while avoiding weird errors and issues.
"""

##########################################################################
## Imports
##########################################################################

from yellowbrick.exceptions import YellowbrickAttributeError


##########################################################################
## Module Constants
##########################################################################

CLASSIFIER = "classifier"
REGRESSOR = "regressor"
CLUSTERER = "clusterer"
DENSITY_ESTIMATOR = "DensityEstimator"
OUTLIER_DETECTOR = "outlier_detector"


##########################################################################
## Functional API
##########################################################################

def wrap(estimator, estimator_type=None):
    """
    Wrap a third-party estimator that implements portions of the scikit-learn API to
    make it available to Yellowbrick visualizers. If the Yellowbrick visualizer cannot
    succeed, then a sensible error is raised instead.

    Parameters
    ----------
    estimator : object
        The non-sklearn estimator to wrap and use for Visualizers

    estimator_type : str, optional
        One of "classifier", "regressor", "clusterer", "DensityEstimator", or
        "outlier_detector" that allows the contrib estimator to pass the scikit-learn
        ``is_classifier``, etc. functions. If not specified, the _estimator_type attr
        is passed through to the underlying estimator.
    """
    return ContribEstimator(estimator, estimator_type)


def classifier(estimator):
    """
    Wrap a third-party classifier to make it available to Yellowbrick visualizers.

    Parameters
    ----------
    estimator : object
        The non-sklearn classifier to wrap and use for Visualizers
    """
    return wrap(estimator, CLASSIFIER)


def regressor(estimator):
    """
    Wrap a third-party regressor to make it available to Yellowbrick visualizers.

    Parameters
    ----------
    estimator : object
        The non-sklearn regressor to wrap and use for Visualizers
    """
    return wrap(estimator, REGRESSOR)


def clusterer(estimator):
    """
    Wrap a third-party clusterer to make it available to Yellowbrick visualizers.

    Parameters
    ----------
    estimator : object
        The non-sklearn clusterer to wrap and use for Visualizers
    """
    return wrap(estimator, CLUSTERER)


##########################################################################
## ContribEstimator - Third Pary Estimator Wrapper
##########################################################################

class ContribEstimator(object):
    """
    Wraps a third party estimator that implements the sckit-learn API and therefore
    could be used with Yellowbrick but doesn't subclass ``BaseEstimator``. Since there
    are a number of pitfalls, this object provides sensible errors and warnings rather
    than completely blowing up, allowing contrib users to identify issues and fix them,
    smoothing the path to getting third party estimators into the Yellowbrick ecosystem.

    Parameters
    ----------
    estimator : object
        The non-sklearn estimator to wrap and use for Visualizers

    estimator_type : str, optional
        One of "classifier", "regressor", "clusterer", "DensityEstimator", or
        "outlier_detector" that allows the contrib estimator to pass the scikit-learn
        ``is_classifier``, etc. functions. If not specified, the _estimator_type attr
        is passed through to the underlying estimator.
    """

    def __init__(self, estimator, estimator_type=None):
        self.estimator = estimator
        # Do not set estimator type if not specified to allow passthrough
        if estimator_type:
            self._estimator_type = estimator_type

    def __getattr__(self, attr):
        # proxy to the wrapped object
        try:
            return getattr(self.estimator, attr)
        except AttributeError:
            raise YellowbrickAttributeError((
                "estimator is missing the '{}' attribute, which is required for this "
                "visualizer - please see the third party estimators documentation."
            ).format(attr))



# yellowbrick.target.base
# Base classes for target visualizers
#
# Author:  Benjamin Bengfort
# Created: Thu Jul 19 09:25:53 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: base.py [d742c57] benjamin@bengfort.com $

"""
Base classes for target visualizers
"""

##########################################################################
# Imports
##########################################################################

from yellowbrick.base import Visualizer


##########################################################################
# TargetVisualizer Base Class
##########################################################################


class TargetVisualizer(Visualizer):
    """
    The base class for target visualizers, generic enough to support any
    computation on a single vector, y. This Visualizer is based on the
    LabelEncoder in sklearn.preprocessing, which only accepts a target y.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    fig : matplotlib Figure, default: None
        The figure to plot the Visualizer on. If None is passed in the current
        plot will be used (or generated if required).

    kwargs : dict
        Keyword arguments that are passed to the base class
    """

    def fit(self, y):
        """
        Fit the visualizer to the target y. Note that this visualizer breaks
        the standard estimator interface, and therefore cannot be used inside
        of pipelines, but must be used separately; similar to how the
        LabelEncoder is used.
        """
        raise NotImplementedError("target visualizers must implement a fit method")



# yellowbrick.classifier.feature_correlation
# Feature correlation to dependent variable visualizer.
#
# Author    Zijie (ZJ) Poh
# Created:  Wed Jul 29 15:30:40 2018 -0700
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: feature_correlation.py [33aec16] 8103276+zjpoh@users.noreply.github.com $

"""
Feature Correlation to Dependent Variable Visualizer.
"""

##########################################################################
# Imports
##########################################################################

import numpy as np

from yellowbrick.utils import is_dataframe
from yellowbrick.target.base import TargetVisualizer
from yellowbrick.exceptions import YellowbrickValueError, YellowbrickWarning

from sklearn.feature_selection import mutual_info_classif
from sklearn.feature_selection import mutual_info_regression

from scipy.stats import pearsonr

##########################################################################
# Supported Correlation Computations
##########################################################################

CORRELATION_LABELS = {
    "pearson": "Pearson Correlation",
    "mutual_info-regression": "Mutual Information",
    "mutual_info-classification": "Mutual Information",
}

CORRELATION_METHODS = {
    "mutual_info-regression": mutual_info_regression,
    "mutual_info-classification": mutual_info_classif,
}

##########################################################################
# Class Feature Correlation
##########################################################################


class FeatureCorrelation(TargetVisualizer):
    """
    Displays the correlation between features and dependent variables.

    This visualizer can be used side-by-side with
    ``yellowbrick.features.JointPlotVisualizer`` that plots a feature
    against the target and shows the distribution of each via a
    histogram on each axis.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    method : str, default: 'pearson'
        The method to calculate correlation between features and target.
        Options include:

            - 'pearson', which uses ``scipy.stats.pearsonr``
            - 'mutual_info-regression', which uses ``mutual_info-regression``
              from ``sklearn.feature_selection``
            - 'mutual_info-classification', which uses ``mutual_info_classif``
              from ``sklearn.feature_selection``

    labels : list, default: None
        A list of feature names to use. If a DataFrame is passed to fit and
        features is None, feature names are selected as the column names.

    sort : boolean, default: False
        If false, the features are are not sorted in the plot; otherwise
        features are sorted in ascending order of correlation.

    feature_index : list,
        A list of feature index to include in the plot.

    feature_names : list of feature names
        A list of feature names to include in the plot.
        Must have labels or the fitted data is a DataFrame with column names.
        If feature_index is provided, feature_names will be ignored.

    color: string
        Specify color for barchart

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    features_ : np.array
        The feature labels

    scores_ : np.array
        Correlation between features and dependent variable.

    Examples
    --------

    >>> viz = FeatureCorrelation()
    >>> viz.fit(X, y)
    >>> viz.show()
    """

    def __init__(
        self,
        ax=None,
        method="pearson",
        labels=None,
        sort=False,
        feature_index=None,
        feature_names=None,
        color=None,
        **kwargs
    ):
        super(FeatureCorrelation, self).__init__(ax, **kwargs)

        self.correlation_labels = CORRELATION_LABELS
        self.correlation_methods = CORRELATION_METHODS

        if method not in self.correlation_labels:
            raise YellowbrickValueError(
                "Method {} not implement; choose from {}".format(
                    method, ", ".join(self.correlation_labels)
                )
            )

        # Parameters
        self.sort = sort
        self.color = color
        self.method = method
        self.labels = labels
        self.feature_index = feature_index
        self.feature_names = feature_names

    def fit(self, X, y, **kwargs):
        """
        Fits the estimator to calculate feature correlation to
        dependent variable.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        y : ndarray or Series of length n
            An array or series of target or class values

        kwargs : dict
            Keyword arguments passed to the fit method of the estimator.

        Returns
        -------
        self : visualizer
            The fit method must always return self to support pipelines.
        """
        self._create_labels_for_features(X)

        self._select_features_to_plot(X)

        # Calculate Features correlation with target variable
        if self.method == "pearson":
            self.scores_ = np.array(
                [pearsonr(x, y, **kwargs)[0] for x in np.asarray(X).T]
            )
        else:
            self.scores_ = np.array(
                self.correlation_methods[self.method](X, y, **kwargs)
            )

        # If feature indices are given, plot only the given features
        if self.feature_index:
            self.scores_ = self.scores_[self.feature_index]
            self.features_ = self.features_[self.feature_index]

        # Sort features by correlation
        if self.sort:
            sort_idx = np.argsort(self.scores_)
            self.scores_ = self.scores_[sort_idx]
            self.features_ = self.features_[sort_idx]

        self.draw()
        return self

    def draw(self):
        """
        Draws the feature correlation to dependent variable, called from fit.
        """
        pos = np.arange(self.scores_.shape[0]) + 0.5

        self.ax.barh(pos, self.scores_, color=self.color)

        # Set the labels for the bars
        self.ax.set_yticks(pos)
        self.ax.set_yticklabels(self.features_)

        return self.ax

    def finalize(self):
        """
        Finalize the drawing setting labels and title.
        """
        self.set_title("Features correlation with dependent variable")

        self.ax.set_xlabel(self.correlation_labels[self.method])

        self.ax.grid(False, axis="y")

    def _create_labels_for_features(self, X):
        """
        Create labels for the features

        NOTE: this code is duplicated from MultiFeatureVisualizer
        """
        if self.labels is None:
            # Use column names if a dataframe
            if is_dataframe(X):
                self.features_ = np.array(X.columns)
            # Otherwise use the column index as the labels
            else:
                _, ncols = X.shape
                self.features_ = np.arange(0, ncols)
        else:
            self.features_ = np.array(self.labels)

    def _select_features_to_plot(self, X):
        """
        Select features to plot.

        feature_index is always used as the filter and
        if filter_names is supplied, a new feature_index
        is computed from those names.
        """
        if self.feature_index:
            if self.feature_names:
                raise YellowbrickWarning(
                    "Both feature_index and feature_names "
                    "are specified. feature_names is ignored"
                )
            if min(self.feature_index) < 0 or max(self.feature_index) >= X.shape[1]:
                raise YellowbrickValueError("Feature index is out of range")
        elif self.feature_names:
            self.feature_index = []
            features_list = self.features_.tolist()
            for feature_name in self.feature_names:
                try:
                    self.feature_index.append(features_list.index(feature_name))
                except ValueError:
                    raise YellowbrickValueError("{} not in labels".format(feature_name))


##########################################################################
# Quick Method
##########################################################################


def feature_correlation(
    X,
    y,
    ax=None,
    method="pearson",
    labels=None,
    sort=False,
    feature_index=None,
    feature_names=None,
    color=None,
    show=True,
    **kwargs
):
    """
    Displays the correlation between features and dependent variables.

    This visualizer can be used side-by-side with
    yellowbrick.features.JointPlotVisualizer that plots a feature
    against the target and shows the distribution of each via a
    histogram on each axis.

    Parameters
    ----------
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features

    y : ndarray or Series of length n
        An array or series of target or class values

    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    method : str, default: 'pearson'
        The method to calculate correlation between features and target.
        Options include:

            - 'pearson', which uses ``scipy.stats.pearsonr``
            - 'mutual_info-regression', which uses ``mutual_info-regression``
              from ``sklearn.feature_selection``
            - 'mutual_info-classification', which uses ``mutual_info_classif``
              from ``sklearn.feature_selection``

    labels : list, default: None
        A list of feature names to use. If a DataFrame is passed to fit and
        features is None, feature names are selected as the column names.

    sort : boolean, default: False
        If false, the features are are not sorted in the plot; otherwise
        features are sorted in ascending order of correlation.

    feature_index : list,
        A list of feature index to include in the plot.

    feature_names : list of feature names
        A list of feature names to include in the plot.
        Must have labels or the fitted data is a DataFrame with column names.
        If feature_index is provided, feature_names will be ignored.

    color: string
        Specify color for barchart

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Returns
    -------
    visualizer : FeatureCorrelation
        Returns the fitted visualizer.
    """

    # Instantiate the visualizer
    visualizer = FeatureCorrelation(
        ax=ax,
        method=method,
        labels=labels,
        sort=sort,
        color=color,
        feature_index=feature_index,
        feature_names=feature_names,
        **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X, y, **kwargs)
    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer
    return visualizer



# yellowbrick.gridsearch.base
# Base class for grid search visualizers
#
# Author:   Phillip Schafer
# Created:  Sat Feb 3 10:18:33 2018 -0500
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: base.py [03724ed] pbs929@users.noreply.github.com $

"""
Base class for grid search visualizers
"""

##########################################################################
## Imports
##########################################################################

import numpy as np
from ..utils import is_gridsearch
from ..base import ModelVisualizer
from ..exceptions import (
    YellowbrickTypeError,
    YellowbrickKeyError,
    YellowbrickValueError,
)


##########################################################################
## Dimension reduction utility
##########################################################################


def param_projection(cv_results, x_param, y_param, metric="mean_test_score"):
    """
    Projects the grid search results onto 2 dimensions.

    The display value is taken as the max over the non-displayed dimensions.

    Parameters
    ----------
    cv_results : dict
        A dictionary of results from the `GridSearchCV` object's `cv_results_`
        attribute.

    x_param : string
        The name of the parameter to be visualized on the horizontal axis.

    y_param : string
        The name of the parameter to be visualized on the vertical axis.

    metric : string (default 'mean_test_score')
        The field from the grid search's `cv_results` that we want to display.

    Returns
    -------
    unique_x_vals : list
        The parameter values that will be used to label the x axis.

    unique_y_vals: list
        The parameter values that will be used to label the y axis.

    best_scores: 2D numpy array (n_y by n_x)
        Array of scores to be displayed for each parameter value pair.
    """
    # Extract the parameter values and score corresponding to each gridsearch
    # trial.
    # These are masked arrays where the cases where each parameter is
    # non-applicable are masked.
    try:
        x_vals = cv_results["param_" + x_param]
    except KeyError:
        raise YellowbrickKeyError(
            "Parameter '{}' does not exist in the grid "
            "search results".format(x_param)
        )
    try:
        y_vals = cv_results["param_" + y_param]
    except KeyError:
        raise YellowbrickKeyError(
            "Parameter '{}' does not exist in the grid "
            "search results".format(y_param)
        )

    if metric not in cv_results:
        raise YellowbrickKeyError(
            "Metric '{}' does not exist in the grid " "search results".format(metric)
        )

    # Get unique, unmasked values of the two display parameters
    unique_x_vals = sorted(list(set(x_vals.compressed())))
    unique_y_vals = sorted(list(set(y_vals.compressed())))
    n_x = len(unique_x_vals)
    n_y = len(unique_y_vals)

    # Get mapping of each parameter value -> an integer index
    int_mapping_1 = {value: idx for idx, value in enumerate(unique_x_vals)}
    int_mapping_2 = {value: idx for idx, value in enumerate(unique_y_vals)}

    # Translate each gridsearch result to indices on the grid
    idx_x = [int_mapping_1[value] if value else None for value in x_vals]
    idx_y = [int_mapping_2[value] if value else None for value in y_vals]

    # Create an array of all scores for each value of the display parameters.
    # This is a n_x by n_y array of lists with `None` in place of empties
    # (my kingdom for a dataframe...)
    all_scores = [[None for _ in range(n_x)] for _ in range(n_y)]
    for x, y, score in zip(idx_x, idx_y, cv_results[metric]):
        if x is not None and y is not None:
            if all_scores[y][x] is None:
                all_scores[y][x] = []
            all_scores[y][x].append(score)

    # Get a numpy array consisting of the best scores for each parameter pair
    best_scores = np.empty((n_y, n_x))
    for x in range(n_x):
        for y in range(n_y):
            if all_scores[y][x] is None:
                best_scores[y, x] = np.nan
            else:
                try:
                    best_scores[y, x] = max(all_scores[y][x])
                except ValueError:
                    raise YellowbrickValueError(
                        "Cannot display grid search results for metric '{}': "
                        "result values may not all be numeric".format(metric)
                    )

    return unique_x_vals, unique_y_vals, best_scores


##########################################################################
## Base Grid Search Visualizer
##########################################################################


class GridSearchVisualizer(ModelVisualizer):
    def __init__(self, estimator, ax=None, **kwargs):
        """
        Check to see if model is an instance of GridSearchCV.
        Should return an error if it isn't.
        """
        # A bit of type checking
        if not is_gridsearch(estimator):
            raise YellowbrickTypeError("This estimator is not a GridSearchCV instance")

        # Initialize the super method.
        super(GridSearchVisualizer, self).__init__(estimator, ax=ax, **kwargs)

    def param_projection(self, x_param, y_param, metric):
        """
        Projects the grid search results onto 2 dimensions.

        The wrapped GridSearch object is assumed to be fit already.
        The display value is taken as the max over the non-displayed dimensions.

        Parameters
        ----------
        x_param : string
            The name of the parameter to be visualized on the horizontal axis.

        y_param : string
            The name of the parameter to be visualized on the vertical axis.

        metric : string (default 'mean_test_score')
            The field from the grid search's `cv_results` that we want to display.

        Returns
        -------
        unique_x_vals : list
            The parameter values that will be used to label the x axis.

        unique_y_vals: list
            The parameter values that will be used to label the y axis.

        best_scores: 2D numpy array (n_y by n_x)
            Array of scores to be displayed for each parameter value pair.
        """
        return param_projection(self.estimator.cv_results_, x_param, y_param, metric)

    def fit(self, X, y=None, **kwargs):
        """
        Fits the wrapped grid search and calls draw().

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        y : ndarray or Series of length n
            An array or series of target or class values

        kwargs: dict
            Keyword arguments passed to the drawing functionality or to the
            Scikit-Learn API. See visualizer specific details for how to use
            the kwargs to modify the visualization or fitting process.

        Returns
        -------
        self : visualizer
            The fit method must always return self to support pipelines.
        """
        self.estimator.fit(X, y)
        self.draw()
        return self



# yellowbrick.text.umap_vis
# Implements UMAP visualizations of documents in 2D space.
#
# Author:   John Healy
# Created:  Mon Dec 03 14:00:00 2018 -0500
#
# Copyright (C) 2019 The sckit-yb developers
# For license information, see LICENSE.txt
#
# ID: umap_vis.py [73a44e5] jchealy@gmail.com $

"""
Implements UMAP visualizations of documents in 2D space.
"""

##########################################################################
## Imports
##########################################################################

import warnings
import numpy as np

from collections import defaultdict

from yellowbrick.draw import manual_legend
from yellowbrick.text.base import TextVisualizer
from yellowbrick.style.colors import resolve_colors
from yellowbrick.exceptions import YellowbrickValueError

from sklearn.pipeline import Pipeline

try:
    from umap import UMAP
except ImportError:
    UMAP = None
except (RuntimeError, AttributeError):
    UMAP = None
    warnings.warn(
        "Error Importing UMAP.  UMAP does not support python 2.7 on Windows 32 bit."
    )

##########################################################################
## Quick Methods
##########################################################################


def umap(
    X,
    y=None,
    ax=None,
    classes=None,
    colors=None,
    colormap=None,
    alpha=0.7,
    show=True,
    **kwargs
):
    """
    Display a projection of a vectorized corpus in two dimensions using UMAP (Uniform
    Manifold Approximation and Projection), a nonlinear dimensionality reduction method
    that is particularly well suited to embedding in two or three dimensions for
    visualization as a scatter plot. UMAP is a relatively new technique but is often
    used to visualize clusters or groups of data points and their relative proximities.
    It typically is fast, scalable, and can be applied directly to sparse matrices
    eliminating the need to run a ``TruncatedSVD`` as a pre-processing step.

    The current default for UMAP is Euclidean distance. Hellinger distance would be a
    more appropriate distance function to use with CountVectorize data. That will be
    released in a forthcoming version of UMAP. In the meantime cosine distance is likely
    a better text default that Euclidean and can be set using the keyword argument
    ``metric='cosine'``.

    Parameters
    ----------

    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features representing the corpus of
        vectorized documents to visualize with umap.

    y : ndarray or Series of length n
        An optional array or series of target or class values for instances.
        If this is specified, then the points will be colored according to
        their class. Often cluster labels are passed in to color the documents
        in cluster space, so this method is used both for classification and
        clustering methods.

    ax : matplotlib axes
        The axes to plot the figure on.

    classes : list of strings
        The names of the classes in the target, used to create a legend.

    colors : list or tuple of colors
        Specify the colors for each individual class

    colormap : string or matplotlib cmap
        Sequential colormap for continuous target

    alpha : float, default: 0.7
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered points more visible.

    show : bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however
        you cannot call ``plt.savefig`` from this signature, nor
        ``clear_figure``. If False, simply calls ``finalize()``

    kwargs : dict
        Pass any additional keyword arguments to the UMAP transformer.

    -------
    visualizer: UMAPVisualizer
        Returns the fitted, finalized visualizer
    """
    # Instantiate the visualizer
    visualizer = UMAPVisualizer(
        ax=ax, classes=classes, colors=colors, colormap=colormap, alpha=alpha, **kwargs
    )

    # Fit the visualizer (calls draw)
    visualizer.fit(X, y, **kwargs)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer object
    return visualizer


##########################################################################
## UMAPVisualizer
##########################################################################


class UMAPVisualizer(TextVisualizer):
    """
    Display a projection of a vectorized corpus in two dimensions using UMAP (Uniform
    Manifold Approximation and Projection), a nonlinear dimensionality reduction method
    that is particularly well suited to embedding in two or three dimensions for
    visualization as a scatter plot. UMAP is a relatively new technique but is often
    used to visualize clusters or groups of data points and their relative proximities.
    It typically is fast, scalable, and can be applied directly to sparse matrices
    eliminating the need to run a ``TruncatedSVD`` as a pre-processing step.

    The current default for UMAP is Euclidean distance. Hellinger distance would be a
    more appropriate distance function to use with CountVectorize data. That will be
    released in a forthcoming version of UMAP. In the meantime cosine distance is likely
    a better text default that Euclidean and can be set using the keyword argument
    ``metric='cosine'``.

    For more, see https://github.com/lmcinnes/umap

    Parameters
    ----------

    ax : matplotlib axes
        The axes to plot the figure on.

    labels : list of strings
        The names of the classes in the target, used to create a legend.
        Labels must match names of classes in sorted order.

    colors : list or tuple of colors
        Specify the colors for each individual class

    colormap : string or matplotlib cmap
        Sequential colormap for continuous target

    random_state : int, RandomState instance or None, optional, default: None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random. The random state is applied to the preliminary
        decomposition as well as UMAP.

    alpha : float, default: 0.7
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered points more visible.

    kwargs : dict
        Pass any additional keyword arguments to the UMAP transformer.

    Examples
    --------

    >>> model = MyVisualizer(metric='cosine')
    >>> model.fit(X)
    >>> model.show()

    """

    # NOTE: cannot be np.nan
    NULL_CLASS = None

    def __init__(
        self,
        ax=None,
        labels=None,
        classes=None,
        colors=None,
        colormap=None,
        random_state=None,
        alpha=0.7,
        **kwargs
    ):

        if UMAP is None:
            raise YellowbrickValueError(
                (
                    "umap package doesn't seem to be installed."
                    "Please install UMAP via: pip install umap-learn"
                )
            )

        # Visual Parameters
        self.alpha = alpha
        self.labels = labels
        self.colors = colors
        self.colormap = colormap
        self.random_state = random_state

        # Fetch UMAP kwargs from kwargs by popping only keys belonging to UMAP params
        umap_kwargs = {
            key: kwargs.pop(key) for key in UMAP().get_params() if key in kwargs
        }

        # UMAP doesn't require any pre-processing before embedding and thus doesn't
        # require a pipeline.
        self.transformer_ = self.make_transformer(umap_kwargs)

        # Call super at the end so that size and title are set correctly
        super(UMAPVisualizer, self).__init__(ax=ax, **kwargs)

    def make_transformer(self, umap_kwargs={}):
        """
        Creates an internal transformer pipeline to project the data set into
        2D space using UMAP. This method will reset the transformer on the
        class.

        Parameters
        ----------
        umap_kwargs : dict
            Keyword arguments for the internal UMAP transformer

        Returns
        -------
        transformer : Pipeline
            Pipelined transformer for UMAP projections
        """

        # Create the pipeline steps
        steps = []

        # Add the UMAP manifold
        steps.append(
            (
                "umap",
                UMAP(n_components=2, random_state=self.random_state, **umap_kwargs),
            )
        )

        # return the pipeline
        return Pipeline(steps)

    def fit(self, X, y=None, **kwargs):
        """
        The fit method is the primary drawing input for the UMAP projection
        since the visualization requires both X and an optional y value. The
        fit method expects an array of numeric vectors, so text documents must
        be vectorized before passing them to this method.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features representing the corpus of
            vectorized documents to visualize with UMAP.

        y : ndarray or Series of length n
            An optional array or series of target or class values for
            instances. If this is specified, then the points will be colored
            according to their class. Often cluster labels are passed in to
            color the documents in cluster space, so this method is used both
            for classification and clustering methods.

        kwargs : dict
            Pass generic arguments to the drawing method

        Returns
        -------
        self : instance
            Returns the instance of the transformer/visualizer
        """

        # Store the classes we observed in y
        if y is not None:
            self.classes_ = np.unique(y)
        elif y is None and self.labels is not None:
            self.classes_ = np.array([self.labels[0]])
        else:
            self.classes_ = np.array([self.NULL_CLASS])

        # Fit our internal transformer and transform the data.
        vecs = self.transformer_.fit_transform(X)
        self.n_instances_ = vecs.shape[0]

        # Draw the vectors
        self.draw(vecs, y, **kwargs)

        # Fit always returns self.
        return self

    def draw(self, points, target=None, **kwargs):
        """
        Called from the fit method, this method draws the UMAP scatter plot,
        from a set of decomposed points in 2 dimensions. This method also
        accepts a third dimension, target, which is used to specify the colors
        of each of the points. If the target is not specified, then the points
        are plotted as a single cloud to show similar documents.
        """
        # Resolve the labels with the classes
        labels = self.labels if self.labels is not None else self.classes_
        if len(labels) != len(self.classes_):
            raise YellowbrickValueError(
                (
                    "number of supplied labels ({}) does not "
                    "match the number of classes ({})"
                ).format(len(labels), len(self.classes_))
            )

        # Create the color mapping for the labels.
        self.color_values_ = resolve_colors(
            n_colors=len(labels), colormap=self.colormap, colors=self.colors
        )
        colors = dict(zip(labels, self.color_values_))

        # Transform labels into a map of class to label
        labels = dict(zip(self.classes_, labels))

        # Expand the points into vectors of x and y for scatter plotting,
        # assigning them to their label if the label has been passed in.
        # Additionally, filter classes not specified directly by the user.
        series = defaultdict(lambda: {"x": [], "y": []})

        if target is not None:
            for t, point in zip(target, points):
                label = labels[t]
                series[label]["x"].append(point[0])
                series[label]["y"].append(point[1])
        else:
            label = self.classes_[0]
            for x, y in points:
                series[label]["x"].append(x)
                series[label]["y"].append(y)

        # Plot the points
        for label, points in series.items():
            self.ax.scatter(
                points["x"], points["y"], c=colors[label], alpha=self.alpha, label=label
            )

        return self.ax

    def finalize(self, **kwargs):
        """
        Finalize the drawing by adding a title and legend, and removing the
        axes objects that do not convey information about UMAP.
        """
        self.set_title("UMAP Projection of {} Documents".format(self.n_instances_))

        # Remove the ticks
        self.ax.set_yticks([])
        self.ax.set_xticks([])

        # Add the legend outside of the figure box.
        if not all(self.classes_ == np.array([self.NULL_CLASS])):
            box = self.ax.get_position()
            self.ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
            manual_legend(
                self,
                self.classes_,
                self.color_values_,
                loc="center left",
                bbox_to_anchor=(1, 0.5),
            )



# yellowbrick.utils
# Utility functions and helpers for the Yellowbrick library.
#
# Author:   Jason Keung
# Author:   Patrick O'Melveny
# Author:   Benjamin Bengfort
# Author:   Rebecca Bilbro
# Created:  Thu Jun 02 15:33:18 2016 -0500
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [79cd8cf] benjamin@bengfort.com $

"""
Utility functions and helpers for the Yellowbrick library.
"""

##########################################################################
## Imports
##########################################################################

from .helpers import *
from .types import *
from .kneed import *



# yellowbrick.cluster
# Visualizers for Cluster analysis and diagnostics
#
# Author:   Benjamin Bengfort
# Created:  Thu Mar 23 17:26:57 2017 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [241edca] benjamin@bengfort.com $

"""
Visualizers for Cluster analysis and diagnostics, particularly visualizations
related to evaluating Scikit-Learn clustering models.
"""

##########################################################################
## Imports
##########################################################################

## Hoist visualizers into the cluster namespace
from .base import *
from .elbow import *
from .silhouette import *
from .icdm import *



# yellowbrick.datasets.signature
# Performs SHA 256 hashing of a file for dataset archive verification.
#
# Author:  Benjamin Bengfort
# Created: Tue Jul 31 14:18:11 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: signature.py [7082742] benjamin@bengfort.com $

"""
Performs SHA 256 hashing of a file for dataset archive verification.
"""

##########################################################################
## Imports
##########################################################################

import hashlib

##########################################################################
## Signature checking utility
##########################################################################


def sha256sum(path, blocksize=65536):
    """
    Computes the SHA256 signature of a file to verify that the file has not
    been modified in transit and that it is the correct version of the data.
    """
    sig = hashlib.sha256()
    with open(path, "rb") as f:
        buf = f.read(blocksize)
        while len(buf) > 0:
            sig.update(buf)
            buf = f.read(blocksize)
    return sig.hexdigest()



# yellowbrick.text
# Visualizers for text feature analysis and diagnostics.
#
# Author:   Rebecca Bilbro
# Created:  2017-01-20 14:42
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [75d9b20] rebecca.bilbro@bytecubed.com $

"""
Visualizers for text feature analysis and diagnostics.
"""

##########################################################################
## Imports
##########################################################################

from .tsne import TSNEVisualizer, tsne
from .umap_vis import UMAPVisualizer, umap
from .freqdist import FreqDistVisualizer, freqdist
from .postag import PosTagVisualizer
from .dispersion import DispersionPlot, dispersion



# yellowbrick.datasets.loaders
# Dataset loading utilities and primary API to the datasets module.
#
# Author:  Benjamin Bengfort
# Created: Tue Jul 31 13:31:23 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: loaders.py [7082742] benjamin@bengfort.com $

"""
Dataset loading utilities and primary API to the datasets module.
"""

##########################################################################
## Imports
##########################################################################

import os
import json

from .base import Dataset, Corpus

__all__ = [
    "load_concrete",
    "load_energy",
    "load_credit",
    "load_occupancy",
    "load_mushroom",
    "load_hobbies",
    "load_game",
    "load_bikeshare",
    "load_spam",
    "load_walking",
    "load_nfl",
]


##########################################################################
## Links and SHA 256 signature of Yellowbrick hosted datasets
##########################################################################

MANIFEST = os.path.join(os.path.dirname(__file__), "manifest.json")
with open(MANIFEST, "r") as f:
    DATASETS = json.load(f)


##########################################################################
## Specific loading utilities
##########################################################################


def _load_dataset(name, data_home=None, return_dataset=False):
    """
    Load a dataset by name and return specified format.
    """
    info = DATASETS[name]
    data = Dataset(name, data_home=data_home, **info)
    if return_dataset:
        return data
    return data.to_data()


def _load_corpus(name, data_home=None):
    """
    Load a corpus object by name.
    """
    info = DATASETS[name]
    return Corpus(name, data_home=data_home, **info)


def load_concrete(data_home=None, return_dataset=False):
    """
    Loads the concrete multivariate dataset that is well suited to regression
    tasks. The dataset contains 1030 instances and 8 real valued attributes
    with a continuous target.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    return_dataset : bool, default=False
        Return the raw dataset object instead of X and y numpy arrays to
        get access to alternative targets, extra features, content and meta.

    Returns
    -------
    X : array-like with shape (n_instances, n_features) if return_dataset=False
        A pandas DataFrame or numpy array describing the instance features.

    y : array-like with shape (n_instances,) if return_dataset=False
        A pandas Series or numpy array describing the target vector.

    dataset : Dataset instance if return_dataset=True
        The Yellowbrick Dataset object provides an interface to accessing the
        data in a variety of formats as well as associated metadata and content.
    """
    return _load_dataset("concrete", data_home, return_dataset)


def load_energy(data_home=None, return_dataset=False):
    """
    Loads the energy multivariate dataset that is well suited to multi-output
    regression and classification tasks. The dataset contains 768 instances and
    8 real valued attributes with two continous targets.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    return_dataset : bool, default=False
        Return the raw dataset object instead of X and y numpy arrays to
        get access to alternative targets, extra features, content and meta.

    Returns
    -------
    X : array-like with shape (n_instances, n_features) if return_dataset=False
        A pandas DataFrame or numpy array describing the instance features.

    y : array-like with shape (n_instances,) if return_dataset=False
        A pandas Series or numpy array describing the target vector.

    dataset : Dataset instance if return_dataset=True
        The Yellowbrick Dataset object provides an interface to accessing the
        data in a variety of formats as well as associated metadata and content.
    """
    return _load_dataset("energy", data_home, return_dataset)


def load_credit(data_home=None, return_dataset=False):
    """
    Loads the credit multivariate dataset that is well suited to binary
    classification tasks. The dataset contains 30000 instances and 23 integer
    and real value attributes with a discrete target.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    return_dataset : bool, default=False
        Return the raw dataset object instead of X and y numpy arrays to
        get access to alternative targets, extra features, content and meta.

    Returns
    -------
    X : array-like with shape (n_instances, n_features) if return_dataset=False
        A pandas DataFrame or numpy array describing the instance features.

    y : array-like with shape (n_instances,) if return_dataset=False
        A pandas Series or numpy array describing the target vector.

    dataset : Dataset instance if return_dataset=True
        The Yellowbrick Dataset object provides an interface to accessing the
        data in a variety of formats as well as associated metadata and content.
    """
    return _load_dataset("credit", data_home, return_dataset)


def load_occupancy(data_home=None, return_dataset=False):
    """
    Loads the occupancy multivariate, time-series dataset that is well suited
    to binary classification tasks. The dataset contains 20560 instances with
    5 real valued attributes and a discrete target.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    return_dataset : bool, default=False
        Return the raw dataset object instead of X and y numpy arrays to
        get access to alternative targets, extra features, content and meta.

    Returns
    -------
    X : array-like with shape (n_instances, n_features) if return_dataset=False
        A pandas DataFrame or numpy array describing the instance features.

    y : array-like with shape (n_instances,) if return_dataset=False
        A pandas Series or numpy array describing the target vector.

    dataset : Dataset instance if return_dataset=True
        The Yellowbrick Dataset object provides an interface to accessing the
        data in a variety of formats as well as associated metadata and content.
    """
    return _load_dataset("occupancy", data_home, return_dataset)


def load_mushroom(data_home=None, return_dataset=False):
    """
    Loads the mushroom multivariate dataset that is well suited to binary
    classification tasks. The dataset contains 8123 instances with 3
    categorical attributes and a discrete target.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    return_dataset : bool, default=False
        Return the raw dataset object instead of X and y numpy arrays to
        get access to alternative targets, extra features, content and meta.

    Returns
    -------
    X : array-like with shape (n_instances, n_features) if return_dataset=False
        A pandas DataFrame or numpy array describing the instance features.

    y : array-like with shape (n_instances,) if return_dataset=False
        A pandas Series or numpy array describing the target vector.

    dataset : Dataset instance if return_dataset=True
        The Yellowbrick Dataset object provides an interface to accessing the
        data in a variety of formats as well as associated metadata and content.
    """
    return _load_dataset("mushroom", data_home, return_dataset)


def load_hobbies(data_home=None):
    """
    Loads the hobbies text corpus that is well suited to classification,
    clustering, and text analysis tasks. The dataset contains 448 documents in
    5 categories with 7420 paragraphs, 14251 sentences, 288520 words, and a
    vocabulary of 23738.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    Returns
    -------
    dataset : Corpus
        The Yellowbrick Corpus object provides an interface to accessing the
        text documents and metadata associated with the corpus.
    """
    return _load_corpus("hobbies", data_home)


def load_game(data_home=None, return_dataset=False):
    """
    Load the Connect-4 game multivariate and spatial dataset that is well
    suited to multiclass classification tasks. The dataset contains 67557
    instances with 42 categorical attributes and a discrete target.

    Note that the game data is stored with categorical features that need to
    be numerically encoded before use with scikit-learn estimators. We
    recommend the use of the ``sklearn.preprocessing.OneHotEncoder`` for this
    task and to develop a ``Pipeline`` using this dataset.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    return_dataset : bool, default=False
        Return the raw dataset object instead of X and y numpy arrays to
        get access to alternative targets, extra features, content and meta.

    Returns
    -------
    X : array-like with shape (n_instances, n_features) if return_dataset=False
        A pandas DataFrame or numpy array describing the instance features.

    y : array-like with shape (n_instances,) if return_dataset=False
        A pandas Series or numpy array describing the target vector.

    dataset : Dataset instance if return_dataset=True
        The Yellowbrick Dataset object provides an interface to accessing the
        data in a variety of formats as well as associated metadata and content.
    """
    return _load_dataset("game", data_home, return_dataset)


def load_bikeshare(data_home=None, return_dataset=False):
    """
    Loads the bike sharing univariate dataset that is well suited to regression
    tasks. The dataset contains 17379 instances with 12 integer and real valued
    attributes and a continuous target.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    return_dataset : bool, default=False
        Return the raw dataset object instead of X and y numpy arrays to
        get access to alternative targets, extra features, content and meta.

    Returns
    -------
    X : array-like with shape (n_instances, n_features) if return_dataset=False
        A pandas DataFrame or numpy array describing the instance features.

    y : array-like with shape (n_instances,) if return_dataset=False
        A pandas Series or numpy array describing the target vector.

    dataset : Dataset instance if return_dataset=True
        The Yellowbrick Dataset object provides an interface to accessing the
        data in a variety of formats as well as associated metadata and content.
    """
    return _load_dataset("bikeshare", data_home, return_dataset)


def load_spam(data_home=None, return_dataset=False):
    """
    Loads the email spam dataset that is weill suited to binary classification
    and threshold tasks. The dataset contains 4600 instances with 57 integer and
    real valued attributes and a discrete target.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    return_dataset : bool, default=False
        Return the raw dataset object instead of X and y numpy arrays to
        get access to alternative targets, extra features, content and meta.

    Returns
    -------
    X : array-like with shape (n_instances, n_features) if return_dataset=False
        A pandas DataFrame or numpy array describing the instance features.

    y : array-like with shape (n_instances,) if return_dataset=False
        A pandas Series or numpy array describing the target vector.

    dataset : Dataset instance if return_dataset=True
        The Yellowbrick Dataset object provides an interface to accessing the
        data in a variety of formats as well as associated metadata and content.
    """
    return _load_dataset("spam", data_home, return_dataset)


def load_walking(data_home=None, return_dataset=False):
    """
    Loads the walking activity dataset that is weill suited to clustering and
    multi-label classification tasks. The dataset contains multi-variate time
    series data with 149,332 real valued measurements across 22 unique walkers.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    return_dataset : bool, default=False
        Return the raw dataset object instead of X and y numpy arrays to
        get access to alternative targets, extra features, content and meta.

    Returns
    -------
    X : array-like with shape (n_instances, n_features) if return_dataset=False
        A pandas DataFrame or numpy array describing the instance features.

    y : array-like with shape (n_instances,) if return_dataset=False
        A pandas Series or numpy array describing the target vector.

    dataset : Dataset instance if return_dataset=True
        The Yellowbrick Dataset object provides an interface to accessing the
        data in a variety of formats as well as associated metadata and content.
    """
    return _load_dataset("walking", data_home, return_dataset)


def load_nfl(data_home=None, return_dataset=False):
    """
    Loads the football receivers dataset that is well suited to clustering
    tasks. The dataset contains 494 instances with 28 integer, real valued, and
    categorical attributes and a discrete target.

    The Yellowbrick datasets are hosted online and when requested, the dataset
    is downloaded to your local computer for use. Note that if the dataset
    hasn't been downloaded before, an Internet connection is required. However,
    if the data is cached locally, no data will be downloaded. Yellowbrick
    checks the known signature of the dataset with the data downloaded to
    ensure the download completes successfully.

    Datasets are stored alongside the code, but the location can be specified
    with the ``data_home`` parameter or the ``$YELLOWBRICK_DATA`` envvar.

    Parameters
    ----------
    data_home : str, optional
        The path on disk where data is stored. If not passed in, it is looked
        up from ``$YELLOWBRICK_DATA`` or the default returned by ``get_data_home``.

    return_dataset : bool, default=False
        Return the raw dataset object instead of X and y numpy arrays to
        get access to alternative targets, extra features, content and meta.

    Returns
    -------
    X : array-like with shape (n_instances, n_features) if return_dataset=False
        A pandas DataFrame or numpy array describing the instance features.

    y : array-like with shape (n_instances,) if return_dataset=False
        A pandas Series or numpy array describing the target vector.

    dataset : Dataset instance if return_dataset=True
        The Yellowbrick Dataset object provides an interface to accessing the
        data in a variety of formats as well as associated metadata and content.
    """
    return _load_dataset("nfl", data_home, return_dataset)



# yellowbrick.features.radviz
# Implements radviz for feature analysis.
#
# Author:   Benjamin Bengfort
# Created:  Fri Oct 07 13:18:00 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: radviz.py [0f4b236] benjamin@bengfort.com $

"""
Implements radviz for feature analysis.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np
import matplotlib.patches as patches

from yellowbrick.draw import manual_legend
from yellowbrick.utils import is_dataframe
from yellowbrick.utils import nan_warnings
from yellowbrick.features.base import DataVisualizer


##########################################################################
## Static RadViz Visualizer
##########################################################################


class RadialVisualizer(DataVisualizer):
    """
    RadViz is a multivariate data visualization algorithm that plots each
    axis uniformely around the circumference of a circle then plots points on
    the interior of the circle such that the point normalizes its values on
    the axes from the center to each arc.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    features : list, default: None
        a list of feature names to use
        The names of the features specified by the columns of the input dataset.
        This length of this list must match the number of columns in X, otherwise
        an exception will be raised on ``fit()``.

    classes : list, default: None
        a list of class names for the legend
        The class labels for each class in y, ordered by sorted class index. These
        names act as a label encoder for the legend, identifying integer classes
        or renaming string labels. If omitted, the class labels will be taken from
        the unique values in y.

        Note that the length of this list must match the number of unique values in
        y, otherwise an exception is raised. This parameter is only used in the
        discrete target type case and is ignored otherwise.

    colors : list or tuple, default: None
        optional list or tuple of colors to colorize lines
        A single color to plot all instances as or a list of colors to color each
        instance according to its class. If not enough colors per class are
        specified then the colors are treated as a cycle.

    colormap : string or cmap, default: None
        optional string or matplotlib cmap to colorize lines
        The colormap used to create the individual colors. If classes are
        specified the colormap is used to evenly space colors across each class.

    alpha : float, default: 1.0
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered points more visible.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Examples
    --------

    >>> visualizer = RadViz()
    >>> visualizer.fit(X, y)
    >>> visualizer.transform(X)
    >>> visualizer.show()

    Attributes
    ----------
    features_ : ndarray, shape (n_features,)
        The names of the features discovered or used in the visualizer that
        can be used as an index to access or modify data in X. If a user passes
        feature names in, those features are used. Otherwise the columns of a
        DataFrame are used or just simply the indices of the data array.

    classes_ : ndarray, shape (n_classes,)
        The class labels that define the discrete values in the target. Only
        available if the target type is discrete. This is guaranteed to be
        strings even if the classes are a different type.
    """

    def __init__(
        self,
        ax=None,
        features=None,
        classes=None,
        colors=None,
        colormap=None,
        alpha=1.0,
        **kwargs
    ):
        if "target_type" not in kwargs:
            kwargs["target_type"] = "discrete"
        super(RadialVisualizer, self).__init__(
            ax=ax,
            features=features,
            classes=classes,
            colors=colors,
            colormap=colormap,
            **kwargs
        )
        self.alpha = alpha

    @staticmethod
    def normalize(X):
        """
        MinMax normalization to fit a matrix in the space [0,1] by column.
        """
        a = X.min(axis=0)
        b = X.max(axis=0)
        return (X - a[np.newaxis, :]) / ((b - a)[np.newaxis, :])

    def fit(self, X, y=None, **kwargs):
        """
        The fit method is the primary drawing input for the
        visualization since it has both the X and y data required for the
        viz and the transform method does not.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        y : ndarray or Series of length n
            An array or series of target or class values

        kwargs : dict
            Pass generic arguments to the drawing method

        Returns
        -------
        self : instance
            Returns the instance of the transformer/visualizer
        """
        super(RadialVisualizer, self).fit(X, y)
        self.draw(X, y, **kwargs)
        return self

    def draw(self, X, y, **kwargs):
        """
        Called from the fit method, this method creates the radviz canvas and
        draws each instance as a class or target colored point, whose location
        is determined by the feature data set.
        """
        # Convert from dataframe
        if is_dataframe(X):
            X = X.values

        # Clean out nans and warn that the user they aren't plotted
        nan_warnings.warn_if_nans_exist(X)
        X, y = nan_warnings.filter_missing(X, y)

        # Get the shape of the data
        nrows, ncols = X.shape

        # Set the axes limits
        self.ax.set_xlim([-1, 1])
        self.ax.set_ylim([-1, 1])

        # Create a data structure to hold scatter plot representations
        to_plot = {label: [[], []] for label in self.classes_}

        # Compute the arcs around the circumference for each feature axis
        # TODO: make this an independent function for override
        s = np.array(
            [
                (np.cos(t), np.sin(t))
                for t in [2.0 * np.pi * (i / float(ncols)) for i in range(ncols)]
            ]
        )

        # Compute the locations of the scatter plot for each class
        # Normalize the data first to plot along the 0, 1 axis
        for i, row in enumerate(self.normalize(X)):
            row_ = np.repeat(np.expand_dims(row, axis=1), 2, axis=1)
            xy = (s * row_).sum(axis=0) / row.sum()
            label = self._label_encoder[y[i]]

            to_plot[label][0].append(xy[0])
            to_plot[label][1].append(xy[1])

        # Add the scatter plots from the to_plot function
        # TODO: store these plots to add more instances to later
        # TODO: make this a separate function
        for label in self.classes_:
            color = self.get_colors([label])[0]
            self.ax.scatter(
                to_plot[label][0],
                to_plot[label][1],
                color=color,
                label=label,
                alpha=self.alpha,
                **kwargs
            )

        # Add the circular axis path
        # TODO: Make this a seperate function (along with labeling)
        self.ax.add_patch(
            patches.Circle(
                (0.0, 0.0),
                radius=1.0,
                facecolor="none",
                edgecolor="grey",
                linewidth=0.5,
            )
        )

        # Add the feature names
        for xy, name in zip(s, self.features_):
            # Add the patch indicating the location of the axis
            self.ax.add_patch(patches.Circle(xy, radius=0.025, facecolor="#777777"))

            # Add the feature names offset around the axis marker
            if xy[0] < 0.0 and xy[1] < 0.0:
                self.ax.text(
                    xy[0] - 0.025,
                    xy[1] - 0.025,
                    name,
                    ha="right",
                    va="top",
                    size="small",
                )
            elif xy[0] < 0.0 and xy[1] >= 0.0:
                self.ax.text(
                    xy[0] - 0.025,
                    xy[1] + 0.025,
                    name,
                    ha="right",
                    va="bottom",
                    size="small",
                )
            elif xy[0] >= 0.0 and xy[1] < 0.0:
                self.ax.text(
                    xy[0] + 0.025,
                    xy[1] - 0.025,
                    name,
                    ha="left",
                    va="top",
                    size="small",
                )
            elif xy[0] >= 0.0 and xy[1] >= 0.0:
                self.ax.text(
                    xy[0] + 0.025,
                    xy[1] + 0.025,
                    name,
                    ha="left",
                    va="bottom",
                    size="small",
                )

        self.ax.axis("equal")
        return self.ax

    def finalize(self, **kwargs):
        """
        Sets the title and adds a legend. Removes the ticks from the graph to
        make a cleaner visualization.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """
        # Set the title
        self.set_title("RadViz for {} Features".format(len(self.features_)))

        # Remove the ticks from the graph
        self.ax.set_yticks([])
        self.ax.set_xticks([])

        # Add the legend
        colors = self.get_colors(self.classes_)
        manual_legend(self, self.classes_, colors, loc="best")


##########################################################################
## Quick Method
##########################################################################


def radviz(
    X,
    y=None,
    ax=None,
    features=None,
    classes=None,
    colors=None,
    colormap=None,
    alpha=1.0,
    show=True,
    **kwargs
):
    """
    Displays each feature as an axis around a circle surrounding a scatter
    plot whose points are each individual instance.

    This helper function is a quick wrapper to utilize the RadialVisualizer
    (Transformer) for one-off analysis.

    Parameters
    ----------

    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features

    y : ndarray or Series of length n, default:None
        An array or series of target or class values

    ax : matplotlib Axes, default: None
        The axes to plot the figure on.

    features : list of strings, default: None
        The names of the features or columns

    classes : list of strings, default: None
        The names of the classes in the target

    colors : list or tuple of colors, default: None
        Specify the colors for each individual class

    colormap : string or matplotlib cmap, default: None
        Sequential colormap for continuous target

    alpha : float, default: 1.0
        Specify a transparency where 1 is completely opaque and 0 is completely
        transparent. This property makes densely clustered points more visible.

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``
        
    kwargs : dict
        Keyword arguments passed to the visualizer base classes.

    Returns
    -------
    viz : RadViz
        Returns the fitted, finalized visualizer
    """
    # Instantiate the visualizer
    visualizer = RadialVisualizer(
        ax=ax,
        features=features,
        classes=classes,
        colors=colors,
        colormap=colormap,
        alpha=alpha,
        **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X, y, **kwargs)
    visualizer.transform(X)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer object
    return visualizer


# Alias for RadViz
RadViz = RadialVisualizer



# yellowbrick.utils.helpers
# Helper functions and generic utilities for use in Yellowbrick code.
#
# Author:   Benjamin Bengfort
# Author:   Rebecca Bilbro
# Created:  Fri May 19 10:39:30 2017 -0700
#
# Copyright (C) 2019 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: helpers.py [79cd8cf] benjamin@bengfort.com $

"""
Helper functions and generic utilities for use in Yellowbrick code.
"""

##########################################################################
## Imports
##########################################################################

import re
import inspect
import sklearn
import numpy as np

from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from yellowbrick.utils.types import is_estimator
from yellowbrick.exceptions import YellowbrickTypeError
from yellowbrick.contrib.wrapper import ContribEstimator


##########################################################################
## Model and Feature Information
##########################################################################

def is_fitted(estimator):
    """
    In order to ensure that we don't call ``fit`` on an already-fitted model,
    this utility function calls ``predict`` on the estimator, returning ``False``
    if it raises a ``sklearn.exceptions.NotFittedError`` and ``True`` otherwise.

    NOTE: This is the solution proposed to scikit-yb: https://bit.ly/2LWQxZO (see
    also: https://stackoverflow.com/a/39900933/6552250), though it remains unclear
    how it will perform with sklearn-style Estimators and Transformers from other
    3rd party libraries like Keras, XGBoost, etc.
    """
    try:
        estimator.predict(np.zeros((7, 3)))
    except sklearn.exceptions.NotFittedError:
        return False
    except AttributeError:
        # Some clustering models (LDA, PCA, Agglomerative) don't implement ``predict``
        try:
            check_is_fitted(
                estimator,
                [
                    "coef_",
                    "estimator_",
                    "labels_",
                    "n_clusters_",
                    "children_",
                    "components_",
                    "n_components_",
                    "n_iter_",
                    "n_batch_iter_",
                    "explained_variance_",
                    "singular_values_",
                    "mean_",
                ],
                all_or_any=any,
            )
            return True
        except sklearn.exceptions.NotFittedError:
            return False
    except Exception:
        # Assume it's fitted, since ``NotFittedError`` wasn't raised
        return True

    return True


def check_fitted(estimator, is_fitted_by="auto", **kwargs):
    """
    Determines whether or not to check if the model has been fitted, and will return
    ``True`` if so. The ``is_fitted_by`` argument is set to ``'auto'`` by default,
    such that the check leaves it to the ``is_fitted`` helper method to determine if
    a ``NotFitted`` error is raised. However, if the user prefers to override this
    automatic functionality (e.g. if a 3rd party sklearn-like estimator has been used
    that doesn't precisely implement the sklearn API), and ``is_fitted_by`` has been
    set to either ``True`` or ``False``, we assume the user has supplied the necessary
    information about whether or not the model is fit using the Visualizer's
    ``is_fitted`` parameter.

    .. todo:: add other measures for checking if an estimator is fitted e.g. by coefs

    Parameters
    -----------
    estimator : sklearn.Estimator
        The model to check fittedness

    is_fitted_by : bool or str, default: 'auto'
        If bool, that value is returned, otherwise ``is_fitted`` is used to check
        for an exception

    kwargs : dict
        Other optional parameters specific to the ``is_fitted_by`` mechanism.

    Returns
    --------
    is_fitted : bool
        Whether or not the model is already fitted
    """
    if isinstance(is_fitted_by, str) and is_fitted_by.lower() == "auto":
        return is_fitted(estimator)
    return bool(is_fitted_by)


def get_model_name(model):
    """
    Detects the model name for a Scikit-Learn model or pipeline.

    Parameters
    ----------
    model: class or instance
        The object to determine the name for. If the model is an estimator it
        returns the class name; if it is a Pipeline it returns the class name
        of the final transformer or estimator in the Pipeline.

    Returns
    -------
    name : string
        The name of the model or pipeline.
    """
    if not is_estimator(model):
        raise YellowbrickTypeError(
            "Cannot detect the model name for non estimator: '{}'".format(type(model))
        )

    if isinstance(model, Pipeline):
        return get_model_name(model.steps[-1][-1])
    elif isinstance(model, ContribEstimator):
        return model.estimator.__class__.__name__
    else:
        return model.__class__.__name__


def has_ndarray_int_columns(features, X):
    """ Checks if numeric feature columns exist in ndarray """
    _, ncols = X.shape
    if not all(d.isdigit() for d in features if isinstance(d, str)) or not isinstance(
        X, np.ndarray
    ):
        return False
    ndarray_columns = np.arange(0, ncols)
    feature_cols = np.unique([int(d) for d in features])
    return all(np.in1d(feature_cols, ndarray_columns))


# Alias for closer name to isinstance and issubclass
hasndarrayintcolumns = has_ndarray_int_columns


def is_monotonic(a, increasing=True):
    """
    Tests whether a vector a has monotonicity.

    Parameters
    ----------
    a : array-like
        Array that should be tested for monotonicity

    increasing : bool, default: True
        Test if the array is montonically increasing, otherwise test if the
        array is montonically decreasing.
    """
    a = np.asarray(a)  # ensure a is array-like

    if a.ndim > 1:
        raise ValueError("not supported for multi-dimensonal arrays")

    if len(a) <= 1:
        return True

    if increasing:
        return np.all(a[1:] >= a[:-1], axis=0)
    return np.all(a[1:] <= a[:-1], axis=0)


def get_param_names(method):
    """
    Returns a list of keyword-only parameter names that may be
    passed into method.

    Parameters
    ----------
    method : function
        The method for which to return keyword-only parameters.

    Returns
    -------
    parameters : list
        A list of keyword-only parameter names for method.
    """
    try:
        signature = inspect.signature(method)
    except (ValueError, TypeError) as e:
        raise e

    parameters = [
        p
        for p in signature.parameters.values()
        if p.name != "self" and p.kind != p.VAR_KEYWORD
    ]

    return sorted([p.name for p in parameters])


##########################################################################
## Numeric Computations
##########################################################################

# From here: https://bit.ly/2xR64lI
def div_safe(numerator, denominator):
    """
    Ufunc-extension that returns 0 instead of nan when dividing numpy arrays

    Parameters
    ----------
    numerator: array-like

    denominator: scalar or array-like that can be validly divided by the numerator

    returns a numpy array

    example: div_safe( [-1, 0, 1], 0 ) == [0, 0, 0]
    """
    # First handle scalars
    if np.isscalar(numerator):
        raise ValueError("div_safe should only be used with an array-like numerator")

    # Then numpy arrays
    try:
        with np.errstate(divide="ignore", invalid="ignore"):
            result = np.true_divide(numerator, denominator)
            result[~np.isfinite(result)] = 0  # -inf inf NaN
        return result
    except ValueError as e:
        raise e


def prop_to_size(vals, mi=0.0, ma=5.0, power=0.5, log=False):
    """
    Converts an array of property values (e.g. a metric or score) to values
    that are more useful for marker sizes, line widths, or other visual
    sizes. The new sizes are computed as:

        y = mi + (ma -mi)(\frac{x_i - min(x){max(x) - min(x)})^{power}

    If ``log=True``, the natural logarithm of the property values is used instead.

    Parameters
    ----------
    prop : array-like, 1D
        An array of values of the property to scale between the size range.

    mi : float, default: 0.0
        The size to assign the smallest property (minimum size value).

    ma : float, default: 5.0
        The size to assign the largest property (maximum size value).

    power : float, default: 0.5
        Used to control how rapidly the size increases from smallest to largest.

    log : bool, default: False
        Use the natural logarithm to compute the property sizes

    Returns
    -------
    sizes : array, 1D
        The new size values, in the same shape as the input prop array
    """
    # ensure that prop is an array
    vals = np.asarray(vals)

    # apply natural log if specified
    if log:
        vals = np.log(vals)

    # avoid division by zero error
    delta = vals.max() - vals.min()
    if delta == 0.0:
        delta = 1.0

    return mi + (ma - mi) * ((vals - vals.min()) / delta) ** power


##########################################################################
# String Computations
##########################################################################


def slugify(text):
    """
    Returns a slug of given text, normalizing unicode data for file-safe
    strings. Used for deciding where to write images to disk.

    Parameters
    ----------
    text : string
        The string to slugify

    Returns
    -------
    slug : string
        A normalized slug representation of the text

    .. seealso:: https://bit.ly/2NW7s1j
    """
    slug = re.sub(r"[^\w]+", " ", text)
    slug = "-".join(slug.lower().strip().split())
    return slug



# yellowbrick.classifier.class_balance
# Class balance visualizer for showing per-class support.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Author:   Neal Humphrey
# Created:  Wed May 18 12:39:40 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: class_balance.py [5388065] neal@nhumphrey.com $

"""
Class balance visualizer for showing per-class support.
"""

##########################################################################
# Imports
##########################################################################

import numpy as np

from yellowbrick.style.colors import resolve_colors
from yellowbrick.target.base import TargetVisualizer
from yellowbrick.exceptions import YellowbrickValueError

from sklearn.utils.multiclass import unique_labels, type_of_target


# Class Balance Modes
BALANCE = "balance"
COMPARE = "compare"


##########################################################################
# Class Balance Chart
##########################################################################


class ClassBalance(TargetVisualizer):
    """
    One of the biggest challenges for classification models is an imbalance of
    classes in the training data. The ClassBalance visualizer shows the
    relationship of the support for each class in both the training and test
    data by displaying how frequently each class occurs as a bar graph.

    The ClassBalance visualizer can be displayed in two modes:

    1. Balance mode: show the frequency of each class in the dataset.
    2. Compare mode: show the relationship of support in train and test data.

    These modes are determined by what is passed to the ``fit()`` method.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    labels: list, optional
        A list of class names for the x-axis if the target is already encoded.
        Ensure that the labels are ordered lexicographically with respect to
        the values in the target. A common use case is to pass
        ``LabelEncoder.classes_`` as this parameter. If not specified, the labels
        in the data will be used.

    colors: list of strings
        Specify colors for the barchart (will override colormap if both are provided).

    colormap : string or matplotlib cmap
        Specify a colormap to color the classes.

    kwargs: dict, optional
        Keyword arguments passed to the super class. Here, used
        to colorize the bars in the histogram.

    Attributes
    ----------
    classes_ : array-like
        The actual unique classes discovered in the target.

    support_ : array of shape (n_classes,) or (2, n_classes)
        A table representing the support of each class in the target. It is a
        vector when in balance mode, or a table with two rows in compare mode.

    Examples
    --------
    To simply observe the balance of classes in the target:

    >>> viz = ClassBalance().fit(y)
    >>> viz.show()

    To compare the relationship between training and test data:

    >>> _, _, y_train, y_test = train_test_split(X, y, test_size=0.2)
    >>> viz = ClassBalance()
    >>> viz.fit(y_train, y_test)
    >>> viz.show()
    """

    def __init__(self, ax=None, labels=None, colors=None, colormap=None, **kwargs):
        self.labels = labels
        self.colors = colors
        self.colormap = colormap

        super(ClassBalance, self).__init__(ax, **kwargs)

    def fit(self, y_train, y_test=None):
        """
        Fit the visualizer to the the target variables, which must be 1D
        vectors containing discrete (classification) data. Fit has two modes:

        1. Balance mode: if only y_train is specified
        2. Compare mode: if both train and test are specified

        In balance mode, the bar chart is displayed with each class as its own
        color. In compare mode, a side-by-side bar chart is displayed colored
        by train or test respectively.

        Parameters
        ----------
        y_train : array-like
            Array or list of shape (n,) that contains discrete data.

        y_test : array-like, optional
            Array or list of shape (m,) that contains discrete data. If
            specified, the bar chart will be drawn in compare mode.
        """

        # check to make sure that y_train is not a 2D array, e.g. X
        if y_train.ndim == 2:
            raise YellowbrickValueError(
                (
                    "fit has changed to only require a 1D array, y "
                    "since version 0.9; please see the docs for more info"
                )
            )

        # Check the target types for the y variables
        self._validate_target(y_train)
        self._validate_target(y_test)

        # Get the unique values from the dataset
        targets = (y_train,) if y_test is None else (y_train, y_test)
        self.classes_ = unique_labels(*targets)

        # Validate the classes with the class names
        if self.labels is not None:
            if len(self.labels) != len(self.classes_):
                raise YellowbrickValueError(
                    (
                        "discovered {} classes in the data, does not match "
                        "the {} labels specified."
                    ).format(len(self.classes_), len(self.labels))
                )

        # Determine if we're in compare or balance mode
        self._mode = BALANCE if y_test is None else COMPARE

        # Compute the support values
        if self._mode == BALANCE:
            self.support_ = np.array([(y_train == idx).sum() for idx in self.classes_])

        else:
            self.support_ = np.array(
                [[(y == idx).sum() for idx in self.classes_] for y in targets]
            )

        # Draw the bar chart
        self.draw()

        # Fit returns self
        return self

    def draw(self):
        """
        Renders the class balance chart on the specified axes from support.
        """
        # Number of colors is either number of classes or 2
        colors = resolve_colors(
            len(self.support_), colormap=self.colormap, colors=self.colors
        )

        if self._mode == BALANCE:
            self.ax.bar(
                np.arange(len(self.support_)),
                self.support_,
                color=colors,
                align="center",
                width=0.5,
            )

        # Compare mode
        else:
            bar_width = 0.35
            labels = ["train", "test"]

            for idx, support in enumerate(self.support_):
                index = np.arange(len(self.classes_))
                if idx > 0:
                    index = index + bar_width

                self.ax.bar(
                    index, support, bar_width, color=colors[idx], label=labels[idx]
                )

        return self.ax

    def finalize(self, **kwargs):
        """
        Finalizes the figure for drawing by setting a title, the legend, and axis
        labels, removing the grid, and making sure the figure is correctly zoomed
        into the bar chart.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """
        # Set the title
        self.set_title("Class Balance for {:,} Instances".format(self.support_.sum()))

        # Set the x ticks with the class names or labels if specified
        labels = self.labels if self.labels is not None else self.classes_
        xticks = np.arange(len(labels))
        if self._mode == COMPARE:
            xticks = xticks + (0.35 / 2)

        self.ax.set_xticks(xticks)
        self.ax.set_xticklabels(labels)

        # Compute the ceiling for the y limit
        cmax = self.support_.max()
        self.ax.set_ylim(0, cmax + cmax * 0.1)
        self.ax.set_ylabel("support")

        # Remove the vertical grid
        self.ax.grid(False, axis="x")

        # Add the legend if in compare mode:
        if self._mode == COMPARE:
            self.ax.legend(frameon=True)

    def _validate_target(self, y):
        """
        Raises a value error if the target is not a classification target.
        """
        # Ignore None values
        if y is None:
            return

        y_type = type_of_target(y)
        if y_type not in ("binary", "multiclass"):
            raise YellowbrickValueError(
                ("'{}' target type not supported, only binary and multiclass").format(
                    y_type
                )
            )


##########################################################################
# Quick Method
##########################################################################


def class_balance(
    y_train,
    y_test=None,
    ax=None,
    labels=None,
    color=None,
    colormap=None,
    show=True,
    **kwargs
):
    """Quick method:

    One of the biggest challenges for classification models is an imbalance of
    classes in the training data. This function vizualizes the relationship of
    the support for each class in both the training and test data by
    displaying how frequently each class occurs as a bar graph.

    The figure can be displayed in two modes:

    1. Balance mode: show the frequency of each class in the dataset.
    2. Compare mode: show the relationship of support in train and test data.

    Balance mode is the default if only y_train is specified. Compare mode
    happens when both y_train and y_test are specified.

    Parameters
    ----------
    y_train : array-like
        Array or list of shape (n,) that containes discrete data.

    y_test : array-like, optional
        Array or list of shape (m,) that contains discrete data. If
        specified, the bar chart will be drawn in compare mode.

    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    labels: list, optional
        A list of class names for the x-axis if the target is already encoded.
        Ensure that the labels are ordered lexicographically with respect to
        the values in the target. A common use case is to pass
        ``LabelEncoder.classes_`` as this parameter. If not specified, the labels
        in the data will be used.

    colors: list of strings
        Specify colors for the barchart (will override colormap if both are provided).

    colormap : string or matplotlib cmap
        Specify a colormap to color the classes.

    show : bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however
        you cannot call ``plt.savefig`` from this signature, nor
        ``clear_figure``. If False, simply calls ``finalize()``

    kwargs: dict, optional
        Keyword arguments passed to the super class. Here, used
        to colorize the bars in the histogram.

    Returns
    -------
    visualizer : ClassBalance
        Returns the fitted visualizer
    """
    # Instantiate the visualizer
    visualizer = ClassBalance(ax=ax, labels=labels, color=None, colormap=None, **kwargs)

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(y_train, y_test)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer
    return visualizer



# yellowbrick.features.rankd
# Implements 1D (histograms) and 2D (joint plot) feature rankings.
#
# Author:   Benjamin Bengfort
# Created:  Fri Oct 07 15:14:01 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: rankd.py [ee754dc] benjamin@bengfort.com $

"""
Implements 1D (histograms) and 2D (joint plot) feature rankings.
"""

##########################################################################
## Imports
##########################################################################

import warnings
import numpy as np
import matplotlib as mpl

from scipy.stats import shapiro
from scipy.stats import spearmanr
from scipy.stats import kendalltau as sp_kendalltau

from yellowbrick.utils import is_dataframe
from yellowbrick.features.base import MultiFeatureVisualizer
from yellowbrick.exceptions import YellowbrickValueError, YellowbrickWarning


__all__ = ["rank1d", "rank2d", "Rank1D", "Rank2D"]


##########################################################################
## Metrics
##########################################################################


def kendalltau(X):
    """
    Accepts a matrix X and returns a correlation matrix so that each column
    is the variable and each row is the observations.

    Parameters
    ----------
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features

    """
    corrs = np.zeros((X.shape[1], X.shape[1]))
    for idx, cola in enumerate(X.T):
        for jdx, colb in enumerate(X.T):
            corrs[idx, jdx] = sp_kendalltau(cola, colb)[0]
    return corrs


##########################################################################
## Base Feature Visualizer
##########################################################################


class RankDBase(MultiFeatureVisualizer):
    """
    Base visualizer for Rank1D and Rank2D

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    fig : matplotlib Figure, default: None
        The figure to plot the Visualizer on. If None is passed in the current
        plot will be used (or generated if required).

    algorithm : string
        The ranking algorithm to use; options and defaults vary by subclass

    features : list
        A list of feature names to use.
        If a DataFrame is passed to fit and features is None, feature
        names are selected as the columns of the DataFrame.

    show_feature_names : boolean, default: True
        If True, the feature names are used to label the axis ticks in the
        plot.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    ranks_ : ndarray
        An n-dimensional, symmetric array of rank scores, where n is the
        number of features. E.g. for 1D ranking, it is (n,), for a
        2D ranking it is (n,n) and so forth.

    Examples
    --------

    >>> visualizer = Rank2D()
    >>> visualizer.fit(X, y)
    >>> visualizer.transform(X)
    >>> visualizer.show()

    Notes
    -----
    These parameters can be influenced later on in the visualization
    process, but can and should be set as early as possible.
    """

    ranking_methods = {}

    def __init__(
        self,
        ax=None,
        fig=None,
        algorithm=None,
        features=None,
        show_feature_names=True,
        **kwargs
    ):
        """
        Initialize the class with the options required to rank and
        order features as well as visualize the result.
        """
        super(RankDBase, self).__init__(ax=ax, fig=fig, features=features, **kwargs)

        # Data Parameters
        self.ranking_ = algorithm

        # Display parameters
        self.show_feature_names_ = show_feature_names

    def transform(self, X, **kwargs):
        """
        The transform method is the primary drawing hook for ranking classes.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        kwargs : dict
            Pass generic arguments to the drawing method

        Returns
        -------
        X : ndarray
            Typically a transformed matrix, X' is returned. However, this
            method performs no transformation on the original data, instead
            simply ranking the features that are in the input data and returns
            the original data, unmodified.
        """
        self.ranks_ = self.rank(X)
        self.draw(**kwargs)

        # Return the X matrix, unchanged
        return X

    def rank(self, X, algorithm=None):
        """
        Returns the feature ranking.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        algorithm : str or None
            The ranking mechanism to use, or None for the default

        Returns
        -------
        ranks : ndarray
            An n-dimensional, symmetric array of rank scores, where n is the
            number of features. E.g. for 1D ranking, it is (n,), for a
            2D ranking it is (n,n) and so forth.
        """
        algorithm = algorithm or self.ranking_
        algorithm = algorithm.lower()

        if algorithm not in self.ranking_methods:
            raise YellowbrickValueError(
                "'{}' is unrecognized ranking method".format(algorithm)
            )

        # Extract matrix from dataframe if necessary
        if is_dataframe(X):
            X = X.values

        return self.ranking_methods[algorithm](X)

    def finalize(self, **kwargs):
        """
        Sets a title on the RankD plot.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """
        # There is a known bug in matplotlib 3.1.1 that affects RankD plots
        # See #912 and #914 for details.
        if mpl.__version__ == "3.1.1":
            msg = (
                "RankD plots may be clipped when using matplotlib v3.1.1, "
                "upgrade to matplotlib v3.1.2 or later to fix the plots."
            )
            warnings.warn(msg, YellowbrickWarning)

        # Set the title for all RankD visualizations.
        self.set_title(
            "{} Ranking of {} Features".format(
                self.ranking_.title(), len(self.features_)
            )
        )


##########################################################################
## Rank 1D Feature Visualizer
##########################################################################


class Rank1D(RankDBase):
    """
    Rank1D computes a score for each feature in the data set with a specific
    metric or algorithm (e.g. Shapiro-Wilk) then returns the features ranked
    as a bar plot.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    algorithm : one of {'shapiro', }, default: 'shapiro'
        The ranking algorithm to use, default is 'Shapiro-Wilk.

    features : list
        A list of feature names to use.
        If a DataFrame is passed to fit and features is None, feature
        names are selected as the columns of the DataFrame.

    orient : 'h' or 'v', default='h'
        Specifies a horizontal or vertical bar chart.

    show_feature_names : boolean, default: True
        If True, the feature names are used to label the x and y ticks in the
        plot.

    color: string
        Specify color for barchart

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    ranks_ : ndarray
        An array of rank scores with shape (n,), where n is the
        number of features. It is computed during `fit`.

    Examples
    --------
    >>> visualizer = Rank1D()
    >>> visualizer.fit(X, y)
    >>> visualizer.transform(X)
    >>> visualizer.show()
    """

    ranking_methods = {"shapiro": lambda X: np.array([shapiro(x)[0] for x in X.T])}

    def __init__(
        self,
        ax=None,
        algorithm="shapiro",
        features=None,
        orient="h",
        show_feature_names=True,
        color=None,
        **kwargs
    ):
        """
        Initialize the class with the options required to rank and
        order features as well as visualize the result.
        """
        super(Rank1D, self).__init__(
            ax=ax,
            algorithm=algorithm,
            features=features,
            show_feature_names=show_feature_names,
            **kwargs
        )
        self.color = color
        self.orientation_ = orient

    def draw(self, **kwargs):
        """
        Draws the bar plot of the ranking array of features.
        """
        if self.orientation_ == "h":
            # Make the plot
            self.ax.barh(np.arange(len(self.ranks_)), self.ranks_, color=self.color)

            # Add ticks and tick labels
            self.ax.set_yticks(np.arange(len(self.ranks_)))
            if self.show_feature_names_:
                self.ax.set_yticklabels(self.features_)
            else:
                self.ax.set_yticklabels([])

            # Order the features from top to bottom on the y axis
            self.ax.invert_yaxis()

            # Turn off y grid lines
            self.ax.yaxis.grid(False)

        elif self.orientation_ == "v":
            # Make the plot
            self.ax.bar(np.arange(len(self.ranks_)), self.ranks_, color=self.color)

            # Add ticks and tick labels
            self.ax.set_xticks(np.arange(len(self.ranks_)))
            if self.show_feature_names_:
                self.ax.set_xticklabels(self.features_, rotation=90)
            else:
                self.ax.set_xticklabels([])

            # Turn off x grid lines
            self.ax.xaxis.grid(False)

        else:
            raise YellowbrickValueError("Orientation must be 'h' or 'v'")


##########################################################################
## Rank 2D Feature Visualizer
##########################################################################


class Rank2D(RankDBase):
    """
    Rank2D performs pairwise comparisons of each feature in the data set with
    a specific metric or algorithm (e.g. Pearson correlation) then returns
    them ranked as a lower left triangle diagram.

    Parameters
    ----------
    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    algorithm : str, default: 'pearson'
        The ranking algorithm to use, one of: 'pearson', 'covariance', 'spearman',
        or 'kendalltau'.

    features : list
        A list of feature names to use.
        If a DataFrame is passed to fit and features is None, feature
        names are selected as the columns of the DataFrame.

    colormap : string or cmap, default: 'RdBu_r'
        optional string or matplotlib cmap to colorize lines
        Use either color to colorize the lines on a per class basis or colormap to
        color them on a continuous scale.

    show_feature_names : boolean, default: True
        If True, the feature names are used to label the axis ticks in the plot.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    ranks_ : ndarray
        An array of rank scores with shape (n,n), where n is the
        number of features. It is computed during `fit`.

    Examples
    --------

    >>> visualizer = Rank2D()
    >>> visualizer.fit(X, y)
    >>> visualizer.transform(X)
    >>> visualizer.show()

    Notes
    -----
    These parameters can be influenced later on in the visualization
    process, but can and should be set as early as possible.
    """

    ranking_methods = {
        "pearson": lambda X: np.corrcoef(X.transpose()),
        "covariance": lambda X: np.cov(X.transpose()),
        "spearman": lambda X: spearmanr(X, axis=0)[0],
        "kendalltau": lambda X: kendalltau(X),
    }

    def __init__(
        self,
        ax=None,
        algorithm="pearson",
        features=None,
        colormap="RdBu_r",
        show_feature_names=True,
        **kwargs
    ):
        """
        Initialize the class with the options required to rank and
        order features as well as visualize the result.
        """
        super(Rank2D, self).__init__(
            ax=ax,
            algorithm=algorithm,
            features=features,
            show_feature_names=show_feature_names,
            **kwargs
        )
        self.colormap = colormap

    def draw(self, **kwargs):
        """
        Draws the heatmap of the ranking matrix of variables.
        """
        # Set the axes aspect to be equal
        self.ax.set_aspect("equal")

        # Generate a mask for the upper triangle
        mask = np.zeros_like(self.ranks_, dtype=bool)
        mask[np.triu_indices_from(mask)] = True

        # Draw the heatmap
        # TODO: Move mesh to a property so the colorbar can be finalized
        data = np.ma.masked_where(mask, self.ranks_)
        mesh = self.ax.pcolormesh(data, cmap=self.colormap, vmin=-1, vmax=1)

        # Set the Axis limits
        self.ax.set(xlim=(0, data.shape[1]), ylim=(0, data.shape[0]))

        # Add the colorbar
        cb = self.ax.figure.colorbar(mesh, None, self.ax)
        cb.outline.set_linewidth(0)

        # Reverse the rows to get the lower left triangle
        self.ax.invert_yaxis()

        # Add ticks and tick labels
        self.ax.set_xticks(np.arange(len(self.ranks_)) + 0.5)
        self.ax.set_yticks(np.arange(len(self.ranks_)) + 0.5)
        if self.show_feature_names_:
            self.ax.set_xticklabels(self.features_, rotation=90)
            self.ax.set_yticklabels(self.features_)
        else:
            self.ax.set_xticklabels([])
            self.ax.set_yticklabels([])


##########################################################################
## Quick Methods
##########################################################################

def rank1d(
    X,
    y=None,
    ax=None,
    algorithm="shapiro",
    features=None,
    orient="h",
    show_feature_names=True,
    color=None,
    show=True,
    **kwargs
):
    """Scores each feature with the algorithm and ranks them in a bar plot.

    This helper function is a quick wrapper to utilize the Rank1D Visualizer
    (Transformer) for one-off analysis.

    Parameters
    ----------
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features

    y : ndarray or Series of length n
        An array or series of target or class values

    ax : matplotlib axes
        the axis to plot the figure on.

    algorithm : one of {'shapiro', }, default: 'shapiro'
        The ranking algorithm to use, default is 'Shapiro-Wilk.

    features : list
        A list of feature names to use.
        If a DataFrame is passed to fit and features is None, feature
        names are selected as the columns of the DataFrame.

    orient : 'h' or 'v'
        Specifies a horizontal or vertical bar chart.

    show_feature_names : boolean, default: True
        If True, the feature names are used to label the axis ticks in the
        plot.

    color: string
        Specify color for barchart

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

     kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Returns
    -------
    viz : Rank1D
        Returns the fitted, finalized visualizer.

    """
    # Instantiate the visualizer
    visualizer = Rank1D(
        ax=ax,
        algorithm=algorithm,
        features=features,
        orient=orient,
        show_feature_names=show_feature_names,
        color=color,
        **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    visualizer.fit(X, y)
    visualizer.transform(X)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer object
    return visualizer


def rank2d(
    X,
    y=None,
    ax=None,
    algorithm="pearson",
    features=None,
    colormap="RdBu_r",
    show_feature_names=True,
    show=True,
    **kwargs
):
    """Rank2D quick method

    Rank2D performs pairwise comparisons of each feature in the data set with
    a specific metric or algorithm (e.g. Pearson correlation) then returns
    them ranked as a lower left triangle diagram.

    Parameters
    ----------
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features to perform the pairwise compairsons on.

    y : ndarray or Series of length n, default: None
        An array or series of target or class values, optional (not used).

    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    algorithm : str, default: 'pearson'
        The ranking algorithm to use, one of: 'pearson', 'covariance', 'spearman',
        or 'kendalltau'.

    features : list
        A list of feature names to use.
        If a DataFrame is passed to fit and features is None, feature names are
        selected as the columns of the DataFrame.

    colormap : string or cmap, default: 'RdBu_r'
        optional string or matplotlib cmap to colorize lines
        Use either color to colorize the lines on a per class basis or colormap to
        color them on a continuous scale.

    show_feature_names : boolean, default: True
        If True, the feature names are used to label the axis ticks in the plot.

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Returns
    -------
    viz : Rank2D
        Returns the fitted, finalized visualizer that created the Rank2D heatmap.
    """
    # Instantiate the visualizer
    viz = Rank2D(
        ax=ax,
        algorithm=algorithm,
        features=features,
        colormap=colormap,
        show_feature_names=show_feature_names,
        **kwargs
    )

    # Fit and transform the visualizer (calls draw)
    viz.fit(X, y)
    viz.transform(X)

    # Show or finalize
    if show:
        viz.show()
    else:
        viz.finalize()

    # Return the visualizer object
    return viz



# yellowbrick.text.dispersion
# Implementations of lexical dispersions for text visualization.
#
# Author:   Larry Gray
# Created:  Fri Jun 22 15:40:49 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: dispersion.py [3822dd6] lwgray@gmail.com $

"""
Implementation of lexical dispersion for text visualization
"""


##########################################################################
## Imports
##########################################################################

import itertools
from collections import defaultdict

import numpy as np

from yellowbrick.text.base import TextVisualizer
from yellowbrick.style.colors import resolve_colors
from yellowbrick.exceptions import YellowbrickValueError


##########################################################################
## Dispersion Plot Visualizer
##########################################################################


class DispersionPlot(TextVisualizer):
    """
    Lexical dispersion illustrates the homogeneity of a word (or set of words) across
    the documents of a corpus.

    DispersionPlot allows for visualization of the lexical dispersion of words in a
    corpus. This plot illustrates with vertical lines the occurrences of one or more
    search terms throughout the corpus, noting how many words relative to the beginning
    of the corpus it appears. If the target vector of the corpus documents is provided,
    the points will be colored with respect to their document category, which allows for
    additional analysis of relationships in search term homogeneity within and across
    document categories. If annotation is requested, document boundaries will be
    displayed as vertical lines in the plot.

    Parameters
    ----------
    search_terms : list
        A list of search terms whose dispersion across a corpus passed at fit
        should be visualized.

    ax : matplotlib axes, default: None
        The axes to plot the figure on.

    colors : list or tuple of colors
        Specify the colors for each individual class. Will override colormap if both are
        provided.

    colormap : string or matplotlib cmap
        Qualitative colormap for discrete target

    ignore_case : boolean, default: False
        Specify whether input will be case-sensitive.

    annotate_docs : boolean, default: False
        Specify whether document boundaries will be displayed. Vertical lines
        are positioned at the end of each document.

    labels : list of strings
        The names of the classes in the target, used to create a legend.
        Labels must match names of classes in sorted order.

    kwargs : dict
        Pass any additional keyword arguments to the super class.

    Attributes
    ----------
    self.classes_ : list
        A list of strings representing the unique classes in the target in sorted order.
        If ``y`` is provided, these are extracted from ``y``, unless a list of class
        labels is provided by the user on instantiation.

    self.boundaries_ : list
        A list of integers indicating the document boundaries with respect to
        word offsets.

    self.indexed_words_ : list
        A list of integers indicating the y position for each occurrence of each of
        the search terms.

    self.word_categories_ : list
        A list of strings indicating the corresponding document category of each search
        term occurrence.
    """

    # NOTE: cannot be np.nan
    NULL_CLASS = None

    def __init__(
        self,
        search_terms,
        ax=None,
        colors=None,
        colormap=None,
        ignore_case=False,
        annotate_docs=False,
        labels=None,
        **kwargs
    ):
        super(DispersionPlot, self).__init__(ax=ax, **kwargs)

        self.labels = labels
        self.colors = colors
        self.colormap = colormap

        self.ignore_case = ignore_case
        self.search_terms = search_terms
        self.annotate_docs = annotate_docs

    def _compute_dispersion(self, X, y):
        """
        Produces a generator containing the offset word count, y_coordinate, and
        label/category for each occurrance of the search terms.

        Attributes
        ----------
        self.boundaries_ : list
            A list of integers indicating the document boundaries with respect to
            word offsets.
        """
        self.boundaries_ = []
        offset = 0

        if y is None:
            y = itertools.repeat(None)

        for doc, category in zip(X, y):
            for word in doc:
                if self.ignore_case:
                    word = word.lower()

                # NOTE: this will find all indices if duplicate words are supplied
                # In the case that word is not in target words, any empty list is
                # returned and no data will be yielded
                offset += 1
                for y_coord in (self.indexed_words_ == word).nonzero()[0]:
                    y_coord = int(y_coord)
                    yield (offset, y_coord, category)

            if self.annotate_docs:
                self.boundaries_.append(offset)

        self.boundaries_ = np.array(self.boundaries_, dtype=int)

    def _check_missing_words(self, points):
        """
        Helper method to raise an error if any of the requested search
        terms do not appear in the corpus.
        """
        for index in range(len(self.indexed_words_)):
            if index in points[:, 1]:
                pass
            else:
                raise YellowbrickValueError(
                    ("The search term '{}' is not found in " "this corpus").format(
                        self.indexed_words_[index]
                    )
                )

    def fit(self, X, y=None, **kwargs):
        """
        The fit method is the primary drawing input for the dispersion
        visualization.

        Parameters
        ----------
        X : list or generator
            Should be provided as a list of documents or a generator
            that yields a list of documents that contain a list of
            words in the order they appear in the document.

        y : ndarray or Series of length n
            An optional array or series of target or class values for
            instances. If this is specified, then the points will be colored
            according to their class.

        Returns
        -------
        self : instance
            Returns the instance of the transformer/visualizer

        Attributes
        ----------
        self.classes_ : list
            A list of strings representing the unique classes in the target in sorted order.
            If ``y`` is provided, these are extracted from ``y``, unless a list of class
            labels is provided by the user on instantiation.

        self.indexed_words_ : list
            A list of integers indicating the y position for each occurrence of each of
            the search terms.

        self.word_categories_ : list
            A list of strings indicating the corresponding document category of each search
            term occurrence.
        """

        if y is not None:
            self.classes_ = np.unique(y)
        else:
            self.classes_ = np.array([self.NULL_CLASS])

        # Create an index (e.g. the y position) for the target words
        self.indexed_words_ = np.flip(self.search_terms, axis=0)
        if self.ignore_case:
            self.indexed_words_ = np.array([w.lower() for w in self.indexed_words_])

        # Stack is used to create a 2D array from the generator
        try:
            offsets_positions_categories = np.stack(self._compute_dispersion(X, y))
        except ValueError:
            raise YellowbrickValueError(("No search terms were found in the corpus"))

        word_positions = np.stack(
            zip(
                offsets_positions_categories[:, 0].astype(int),
                offsets_positions_categories[:, 1].astype(int),
            )
        )

        self.word_categories_ = offsets_positions_categories[:, 2]

        self._check_missing_words(word_positions)

        self.draw(word_positions, **kwargs)
        return self

    def draw(self, points, **kwargs):
        """
        Called from the fit method, this method creates the canvas and
        draws the plot on it.

        Parameters
        ----------
        kwargs: generic keyword arguments.
        """

        # Resolve the labels with the classes
        labels = self.labels if self.labels is not None else self.classes_
        if len(labels) != len(self.classes_):
            raise YellowbrickValueError(
                (
                    "number of supplied labels ({}) does not "
                    "match the number of classes ({})"
                ).format(len(labels), len(self.classes_))
            )

        # Create the color mapping for the labels.
        color_values = resolve_colors(
            n_colors=len(labels), colormap=self.colormap, colors=self.colors
        )
        colors = dict(zip(labels, color_values))

        # Transform labels into a map of class to label
        labels = dict(zip(self.classes_, labels))

        # Define boundaries with a vertical line
        if self.annotate_docs:
            for xcoords in self.boundaries_:
                self.ax.axvline(x=xcoords, color="lightgray", linestyle="dashed")

        series = defaultdict(lambda: {"x": [], "y": []})

        if self.word_categories_ is not None:
            for point, t in zip(points, self.word_categories_):
                label = labels[t]
                series[label]["x"].append(point[0])
                series[label]["y"].append(point[1])
        else:
            label = self.classes_[0]
            for x, y in points:
                series[label]["x"].append(x)
                series[label]["y"].append(y)

        for label, points in series.items():
            self.ax.scatter(
                points["x"],
                points["y"],
                marker="|",
                c=colors[label],
                zorder=100,
                label=label,
            )

        self.ax.set_yticks(list(range(len(self.indexed_words_))))
        self.ax.set_yticklabels(self.indexed_words_)

        return self.ax

    def finalize(self, **kwargs):
        """
        Prepares the figure for rendering by adding a title, axis labels, and
        managing the limits of the text labels. Adds a legend outside of the plot.

        Parameters
        ----------
        kwargs: generic keyword arguments.

        Notes
        -----
        Generally this method is called from show and not directly by the user.
        """
        self.set_title("Lexical Dispersion Plot")
        self.ax.set_ylim(-1, len(self.indexed_words_))
        self.ax.set_xlabel("Word Offset")
        self.ax.grid(False)

        # Add the legend outside of the figure box.
        if not all(self.classes_ == np.array([self.NULL_CLASS])):
            box = self.ax.get_position()
            self.ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
            self.ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))


##########################################################################
## Quick Method
##########################################################################


def dispersion(
    search_terms,
    corpus,
    y=None,
    ax=None,
    colors=None,
    colormap=None,
    annotate_docs=False,
    ignore_case=False,
    labels=None,
    show=True,
    **kwargs
):
    """ Displays lexical dispersion plot for words in a corpus

    This helper function is a quick wrapper to utilize the DispersionPlot
    Visualizer for one-off analysis

    Parameters
    ----------

    search_terms : list
        A list of words whose dispersion will be examined within a corpus

    corpus : list
        Should be provided as a list of documents that contain
        a list of words in the order they appear in the document.

    y : ndarray or Series of length n
        An optional array or series of target or class values for
        instances. If this is specified, then the points will be colored
        according to their class.

    ax : matplotlib axes, default: None
        The axes to plot the figure on.

    colors : list or tuple of colors
        Specify the colors for each individual class. Will override colormap if both are
        provided.

    colormap : string or matplotlib cmap
        Qualitative colormap for discrete target

    annotate_docs : boolean, default: False
        Specify whether document boundaries will be displayed.  Vertical lines
        are positioned at the end of each document.

    ignore_case : boolean, default: False
        Specify whether input  will be case-sensitive.

    labels : list of strings
        The names of the classes in the target, used to create a legend.
        Labels must match names of classes in sorted order.

    show : bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however
        you cannot call ``plt.savefig`` from this signature, nor
        ``clear_figure``. If False, simply calls ``finalize()``

    kwargs : dict
        Pass any additional keyword arguments to the super class.

    Returns
    -------
    viz: DispersionPlot
        Returns the fitted, finalized visualizer
    """

    # Instantiate the visualizer
    visualizer = DispersionPlot(
        search_terms,
        ax=ax,
        colors=colors,
        colormap=colormap,
        ignore_case=ignore_case,
        labels=labels,
        annotate_docs=annotate_docs,
        **kwargs
    )

    visualizer.fit(corpus, y, **kwargs)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    return visualizer



# yellowbrick.gridsearch
# Visualizers for the results of GridSearchCV.
#
# Author:   Phillip Schafer
# Created:  Sat Feb 3 10:18:33 2018 -0500
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [03724ed] pbs929@users.noreply.github.com $

"""
Visualizers for the results of GridSearchCV.
"""

##########################################################################
## Imports
##########################################################################

## Hoist visualizers into the gridsearch namespace
from .pcolor import *
